import os

# Determine the correct config directory based on OS
def get_config_directory():
    if os.name == "nt":  # Windows
        return os.path.join(os.getenv("APPDATA"), "bluebook")
    else:  # macOS/Linux
        return os.path.join(os.path.expanduser("~"), ".config", "bluebook")

# Ensuring config directory exists
os.makedirs(get_config_directory(), exist_ok=True)

class Configuration:

    class SystemPath:
           CONFIG_DIR = get_config_directory()
           CONFIG_PATH = os.path.join(CONFIG_DIR, "config.json")
           DATABASE_PATH =os.path.join(CONFIG_DIR, "storage.db")



from sqlmodel import Field, SQLModel, Session, UniqueConstraint, create_engine, select, delete
from bluebook.confguration import Configuration


# Data Models
class ExtraRequest(SQLModel, table=True):
    __table_args__ = (UniqueConstraint("request"),)
    id: int | None = Field(default=None, primary_key=True)
    request: str 

    def to_dict(self):
        return {'id': self.id, 'request': self.request}


class Database:
    def __init__(self):
        # Setup the database
        self.engine = create_engine(f"sqlite:///{Configuration.SystemPath.DATABASE_PATH}")
        SQLModel.metadata.create_all(self.engine)

    def select_all_extra_requests(self):
        with Session(self.engine) as session:
            return session.exec(select(ExtraRequest)).all()
    
    def select_extra_req_by_id(self, id: int | str):
        if type(id) is str:
            try:
                id = int(id) # Best effort to convert to int
            except:
                pass
        with Session(self.engine) as session:
            return session.exec(select(ExtraRequest).where(ExtraRequest.id==id)).first()

    def select_extra_req_by_value(self, request: str):
        with Session(self.engine) as session:
            return session.exec(select(ExtraRequest).where(ExtraRequest.request == request)).first()
    
    def add_extra_request(self, request: int):
        extra_request = ExtraRequest(request=request)
        with Session(self.engine) as session:
            session.add(extra_request)
            session.commit()
    
    def remove_extra_request_by_id(self, id):
        if type(id) is str:
            try:
                id = int(id) # Best effort to convert to int
            except:
                pass
        with Session(self.engine) as session:
            session.exec(delete(ExtraRequest).where(ExtraRequest.id==id))
            session.commit()
    
    def remove_extra_request_by_value(self, request):
        with Session(self.engine) as session:
            session.exec(delete(ExtraRequest).where(ExtraRequest.request==request))
            session.commit()
        


from flask import Flask, render_template, request, session, redirect
from logging.config import dictConfig
import google.genai.errors
import os
import random
import click
import sqlalchemy.exc
from bluebook import generator
from bluebook import token_manager
from bluebook import database_manager

class Statistics:
    def __init__(self):
        self.all_num = 0
        self.correct = 0
    
    def get_correct_num(self):
        return self.correct
    
    def get_incorrect_num(self):
        return self.all_num - self.correct
    
    def increment_correct(self):
        self.correct += 1

    def increment_all_num(self):
        self.all_num += 1

    def increment_both(self):
        self.increment_all_num()
        self.increment_correct()

    def serialise(self):
        return {"all": self.all_num, "correct": self.correct, "incorrect": self.get_incorrect_num()}


# Compute the directory of the current file
app_dir = os.path.dirname(os.path.abspath(__file__))

# Set the absolute paths for templates and static folders
template_dir = os.path.join(app_dir, 'templates')
static_dir = os.path.join(app_dir, 'static')


# Initialize the application and its state
app = Flask("blue-book", template_folder=template_dir, static_folder=static_dir)
state: list[generator.Question] = [] # Essentially a list of gennerated questions
app.secret_key = random.randbytes(32)
db_manager = database_manager.Database()


def set_additional_request(value):
    if not value:
        session['additional_request'] = {'set': False, 'value': '', 'saved': False}
    else:
        saved_request = db_manager.select_extra_req_by_value(value)
        if saved_request:
            session['additional_request'] = {'set': True, 'value': value, 'saved': True}
        else:
            session['additional_request'] = {'set': True, 'value': value, 'saved': False}


def ensure_session():
    if 'submitted' not in session:
        session['submitted'] = False
    if 'additional_request' not in session:
        set_additional_request(False)
    if 'latest_num' not in session:
        session['latest_num'] = '2'
    if 'TOKEN_PRESENT' not in session:
        session['TOKEN_PRESENT'] = False


def obtain_saved_topics():
    data = {}
    all_saved_topics = db_manager.select_all_extra_requests()
    size = len(all_saved_topics)
    data['size'] = size
    data['requests'] = []
    for topic in all_saved_topics:
        data['requests'].append(topic.to_dict())
    return data


@app.route("/generate", methods=['POST'])
def generate():
    config = token_manager.load_config()
    ensure_session()
    if token_page:= token_manager.ensure_token(config):
        return token_page
    session['submitted'] = True
    num_of_questions = int(request.form['num_of_questions'])
    session['latest_num'] = str(num_of_questions)
    additional_request = generator.sanitise_input(str(request.form['additional_request']))
    if "additional_request_preset" in request.form:
        if request.form['additional_request_preset']:
            additional_request = generator.sanitise_input(str(request.form['additional_request_preset']))
    if not additional_request:
        app.logger.debug(f"Generating {num_of_questions} new questions")
        set_additional_request(False)
    else:
        app.logger.debug(f"Generating {num_of_questions} new questions with additional request {additional_request}")
        set_additional_request(additional_request)
    try:
        gemini_response = generator.ask_gemini(num_of_questions, config['API_TOKEN'], additional_request=additional_request)
    except google.genai.errors.ClientError:
        return render_template("token_prompt.html.j2")
    global state
    state = gemini_response
    return root()


@app.route("/")
def root():
    config = token_manager.load_config()
    ensure_session()
    global state
    serialized_state = generator.serialize_questions(question_list=state)
    if not serialized_state:
        serialized_state['size'] = 0
    if token_manager.is_token_present(config):
        session['TOKEN_PRESENT'] = True
    else:
        session['TOKEN_PRESENT'] = False
    #app.logger.debug(serialized_state)
    return render_template("root.html.j2", data=serialized_state, saved_topics=obtain_saved_topics())


@app.route("/save_token", methods=["POST"])
def save_token():
    api_token = request.form.get("API_TOKEN")
    config = token_manager.load_config()
    config["API_TOKEN"] = api_token
    token_manager.save_config(config)
    return root()


@app.route("/clear_token", methods=["POST"])
def clear_token():
    token_manager.clear_token()
    return root()


@app.route("/wipe_questions", methods=["POST"])
def wipe_questions():
    session['submitted'] = False
    set_additional_request(False)
    session['latest_num'] = '2'
    global state
    state = []
    if 'TOKEN_PRESENT' not in session:
        session['TOKEN_PRESENT'] = False
    return root()


@app.route("/check", methods=["POST"])
def check():
    ensure_session()
    user_answers = {key: request.form[key] for key in request.form}
    app.logger.debug(user_answers)
    global state
    original_data = state
    statistics = Statistics()
    data_out = {"original_data": generator.serialize_questions(original_data), "user_answers": {}, "is_answer_correct":{}, "statistics": {}}
    for i in range(len(original_data)):
        if original_data[i].choices[int(user_answers[str(i)])].is_correct:
            app.logger.debug(f"Question {i} Correct!")
            data_out["user_answers"][i] = int(user_answers[str(i)])
            data_out["is_answer_correct"][i] = True
            statistics.increment_both()
        else:
            app.logger.debug(f"Question {i} Incorrect!")
            data_out["user_answers"][i] = int(user_answers[str(i)])
            data_out["is_answer_correct"][i] = False
            statistics.increment_all_num()
    data_out['statistics'] = statistics.serialise()
    app.logger.debug(data_out)
    return render_template("check.html.j2", data=data_out, saved_topics=obtain_saved_topics())


@app.route('/save-the-topic', methods=['POST'])
def save_the_topic():
    ensure_session()
    if "topic" in request.form:
        topic_to_save = session['additional_request']['value']
        try:
            db_manager.add_extra_request(topic_to_save)
            set_additional_request(topic_to_save) # To update session
        except sqlalchemy.exc.IntegrityError:
            app.logger.info("Topic was NOT saved: Already present.")
            pass
    return redirect("/")


@app.route('/remove-saved-topic', methods=['POST'])
def remove_saved_topic():
    ensure_session()
    if "topic" in request.form:
        topic_to_delete = session['additional_request']['value']
        if db_manager.select_extra_req_by_value(topic_to_delete):
            app.logger.debug(f'Attempting to delete saved topic: {topic_to_delete}')
            db_manager.remove_extra_request_by_value(topic_to_delete)
            app.logger.info(f'Topic was removed: {topic_to_delete}')
            set_additional_request(topic_to_delete) # To update session
    return redirect("/")


@click.group()
def bluebook():
    '''
    Blue Book - simple CompTIA Sec+ questions generator. Based on gemini-flash-lite model
    '''
    pass


@bluebook.command()
@click.option("--debug", is_flag=True, show_default=True, default=False, help="Run flask app in debug mode")
def start(debug):
    '''
    Start web server
    '''
    if debug:
        dictConfig({
            'version': 1,
            'formatters': {'default': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            }},
            'handlers': {'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default'
            }},
            'root': {
                'level': 'INFO',
                'handlers': ['wsgi']
            }
        })
        app.run("0.0.0.0", "5000", True, True)
    else:
        dictConfig({
            'version': 1,
            'formatters': {'default': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            }},
            'handlers': {'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default'
            }},
            'root': {
                'level': 'INFO',
                'handlers': ['wsgi']
            }
        })
        app.run("0.0.0.0", "5000", False, True)


if __name__ == "__main__":
    bluebook()


import os
import json
import logging
from flask import render_template
from bluebook.confguration import Configuration

# Initialize the logger
logger = logging.getLogger(__name__)


# Function to load configuration
def load_config():
    if os.path.exists(Configuration.SystemPath.CONFIG_PATH):
        with open(Configuration.SystemPath.CONFIG_PATH, "r") as f:
            logger.debug(f'Config has been read from {Configuration.SystemPath.CONFIG_PATH}')
            return json.load(f)
        logger.info(f'Config is empty or not present.')
    return {}


# Function to save configuration
def save_config(config):
    with open(Configuration.SystemPath.CONFIG_PATH, "w") as f:
        json.dump(config, f, indent=4)
    logger.info(f'Config has been saved into {Configuration.SystemPath.CONFIG_PATH}')


def is_token_present(config):
    if "API_TOKEN" not in config:
        logger.debug(f'API TOKEN has not been found in {Configuration.SystemPath.CONFIG_PATH}')
        return False
    elif config["API_TOKEN"] == "":
        logger.debug(f'API TOKEN is empty in {Configuration.SystemPath.CONFIG_PATH}')
        return False
    else:
        logger.debug(f'API TOKEN found in {Configuration.SystemPath.CONFIG_PATH}')
        return True


# Function to ensure the API token is present
def ensure_token(config):
    if not is_token_present(config):
        return render_template("token_prompt.html.j2")
    return None


# Function to clear the API token
def clear_token():
    if os.path.exists(Configuration.SystemPath.CONFIG_PATH):
        with open(Configuration.SystemPath.CONFIG_PATH, "w") as f:
            json.dump({"API_TOKEN": ""}, f, indent=4)
    logger.info(f'API TOKEN has been cleared in {Configuration.SystemPath.CONFIG_PATH}')


from google import genai
from pydantic import BaseModel
import logging
import bleach
import re

logger = logging.getLogger("[Generator]")

class Choice(BaseModel):
    option: str
    is_correct: bool
    explanation: str

    def escape(self):
        self.option = bleach.clean(self.option)
        self.explanation = bleach.clean(self.explanation)


class Question(BaseModel):
    question: str
    choices: list[Choice]

    def escape(self):
        self.question = bleach.clean(self.question)
        for choice in self.choices:
            choice.escape()


def sanitise_input(input: str):
    sanitized = ''
    if len(input) > 90:
        sanitized = re.sub('[^0-9a-zA-Z ]+', '', input[:90])
        sanitized = bleach.clean(sanitized)
    else:
        sanitized = re.sub('[^0-9a-zA-Z ]+', '', input)
        sanitized = bleach.clean(sanitized)
    return sanitized


def serialize_questions(question_list):
    serialized = {"questions": [], "size":0}
    for question in question_list:
        serialized['questions'].append({'question': question.question, 'choices':[]})
        for choice in question.choices:
            serialized['questions'][-1]['choices'].append({'option': choice.option, 'is_correct': choice.is_correct, 'explanation': choice.explanation})
        serialized['size'] += 1
    return serialized


TEST_RESPONSE = {'questions': [{'question': 'A security administrator is tasked with implementing a new authentication system. The primary requirement is to ensure that users can only access resources during specific times of the day and on specific days of the week. Which of the following authentication methods would BEST satisfy this requirement?', 'choices': [{'option': 'Multi-factor authentication with biometric login.', 'is_correct': False, 'explanation': "While multi-factor authentication enhances security, biometrics alone doesn't inherently incorporate time-based restrictions. It verifies identity, not access schedules."}, {'option': 'Role-Based Access Control (RBAC) with time-based restrictions applied to the roles.', 'is_correct': True, 'explanation': 'RBAC allows administrators to assign permissions based on user roles. Time-based restrictions can be implemented to limit access to specific resources or actions during certain times or days, directly addressing the requirements.'}, {'option': 'Single Sign-On (SSO) with password synchronization across all applications.', 'is_correct': False, 'explanation': "SSO streamlines authentication, but it doesn't inherently provide time-based access controls. Password synchronization improves user experience but doesn't restrict access by time."}, {'option': 'Attribute-Based Access Control (ABAC) with location-based and device-based authentication.', 'is_correct': False, 'explanation': 'ABAC is more complex than RBAC. Although location and device can be attributes, the question highlights the time-based requirement. Without time configured, ABAC does not satisfy this primary requirement.'}]}, {'question': "A penetration tester has successfully gained access to a company's internal network by exploiting a vulnerability in a web application. After establishing a foothold, the tester wants to maintain persistent access and avoid detection. Which of the following techniques would be MOST effective for achieving this goal?", 'choices': [{'option': 'Conducting a full system audit and reporting all identified vulnerabilities immediately.', 'is_correct': False, 'explanation': "This action would alert the security team of the penetration testing activity and compromise the tester's access."}, {'option': 'Installing a rootkit to hide malicious processes and backdoors, along with creating a scheduled task to re-establish access if the initial foothold is lost.', 'is_correct': True, 'explanation': 'Rootkits provide persistent, stealthy access by hiding malicious code and processes. A scheduled task ensures that access is re-established even if the initial entry point is patched or removed. This combination is highly effective for maintaining long-term access.'}, {'option': 'Deleting all log files and system audit trails to cover their tracks.', 'is_correct': False, 'explanation': "While deleting logs can help hide activity, it's often detectable and doesn't establish persistent access. This is a reactive, not proactive, approach to persistence."}, {'option': "Changing the user's password to prevent other users from accessing the system and reporting the change to IT support to receive administrator privileges.", 'is_correct': False, 'explanation': 'This is an extremely aggressive approach that would likely trigger immediate security alerts and disrupt normal system operations, making it easily detectable and not ideal for stealthy persistence.'}]}], "size": 2}


def gen_query(question_num, additional_request):
    query = f"""You are a Comptia Security+ examiner with 10 years of experience. You are the best in the world in creating preparation test questions for security+.
You need to create {str(question_num)} multiple choice questions. You are passionate about your work and aim to create detailed questions that are as close to exam questions as possible. Your explanations are in depth and allow students not only to check their knwoledge with your tests, but also learn something they did not know before.  """
    if additional_request:
        query += f"The student has specifically asked you to focus on following topic(s): '{additional_request}'. You must focus on the mentioned topic and create questions related to it or to its adjacent Comptia Security+ Exam objective."
    return query


def ask_gemini(question_num, token, additional_request):
    query = gen_query(question_num, additional_request)
    client = genai.Client(api_key=token)
    try:
        response = client.models.generate_content(
            model="gemini-2.0-flash-lite-preview-02-05",
            contents=query,
            config={
                'response_mime_type': 'application/json',
                'response_schema': list[Question],
            },
        )
    # if server error, return empty list
    except genai.errors.ServerError as e:
        logger.error(f"Client error: {e}")
        return []

    logger.debug(f"Response: {response.text}")
    questions: list[Question] = response.parsed
    for question in questions:
        question.escape()
    return questions




if __name__ == "__main__":
    print(ask_gemini(2))